from __future__ import print_function
from __future__ import unicode_literals


base_interfaces = {
  "ATM": "ATM",
  "AT": "ATM",
  "EOBC": "EOBC",
  "EO": "EOBC",
  "Ethernet": "Ethernet",
  "Eth": "Ethernet",
  "Et": "Ethernet",
  "FastEthernet": "FastEthernet",
  "FastEth": "FastEthernet",
  "FastE": "FastEthernet",
  "Fast": "FastEthernet",
  "Fas": "FastEthernet",
  "FE": "FastEthernet",
  "Fa": "FastEthernet",
  "Fddi": "Fddi",
  "FD": "Fddi",
  "FortyGigabitEthernet": "FortyGigabitEthernet",
  "FortyGigEthernet": "FortyGigabitEthernet",
  "FortyGigEth": "FortyGigabitEthernet",
  "FortyGigE": "FortyGigabitEthernet",
  "FortyGig": "FortyGigabitEthernet",
  "FGE": "FortyGigabitEthernet",
  "FO": "FortyGigabitEthernet",
  "Fo": "FortyGigabitEthernet",
  "GigabitEthernet": "GigabitEthernet",
  "GigEthernet": "GigabitEthernet",
  "GigEth": "GigabitEthernet",
  "GigE": "GigabitEthernet",
  "Gig": "GigabitEthernet",
  "GE": "GigabitEthernet",
  "Gi": "GigabitEthernet",
  "HundredGigabitEthernet": "HundredGigabitEthernet",
  "HundredGigEthernet": "HundredGigabitEthernet",
  "HundredGigEth": "HundredGigabitEthernet",
  "HundredGigE": "HundredGigabitEthernet",
  "HundredGig": "HundredGigabitEthernet",
  "Hu": "HundredGigabitEthernet",
  "Loopback": "Loopback",
  "Lo": "Loopback",
  "lo": "Loopback",
  "Management": "Management",
  "Mgmt": "Management",
  "Ma": "Management",
  "Management_short": "Ma",
  "MFR": "MFR",
  "Multilink": "Multilink",
  "Mu": "Multilink",
  "PortChannel": "PortChannel",
  "Port-Channel": "PortChannel",
  "Po": "PortChannel",
  "POS": "POS",
  "PO": "POS",
  "Serial": "Serial",
  "Se": "Serial",
  "S": "Serial",
  "TenGigabitEthernet": "TenGigabitEthernet",
  "TenGigEthernet": "TenGigabitEthernet",
  "TenGigEth": "TenGigabitEthernet",
  "TenGig": "TenGigabitEthernet",
  "TeGig": "TenGigabitEthernet",
  "Ten": "TenGigabitEthernet",
  "T": "TenGigabitEthernet",
  "Te": "TenGigabitEthernet",
  "Tunnel": "Tunnel",
  "Tun": "Tunnel",
  "Tu": "Tunnel",
  "Virtual-Access": "Virtual-Access",
  "Vi": "Virtual-Access",
  "Virtual-Template": "Virtual-Template",
  "Vt": "Virtual-Template",
  "VLAN": "VLAN",
  "V": "VLAN",
  "Vl": "VLAN"
}

reverse_mapping = {
  "ATM": "At",
  "EOBC": "EO",
  "Ethernet": "Et",
  "FastEthernet": "Fa",
  "Fddi": "FD",
  "FortyGigabitEthernet": "Fo",
  "GigabitEthernet": "Gi",
  "HundredGigabitEthernet": "Hu",
  "Loopback": "Lo",
  "Management": "Ma",
  "MFR": "MFR",
  "Multilink": "Mu",
  "PortChannel": "Po",
  "POS": "PO",
  "Serial": "Se",
  "TenGigabitEthernet": "Te",
  "Tunnel": "Tu",
  "Virtual-Access": "Vi",
  "Virtual-Template": "Vt",
  "VLAN": "Vl"
}
